r"""Functional interface"""

import torch
import pydec
from torch import Tensor
import torch.nn.functional as F
from .._composition import Composition, IndexComposition
from pydec.core.decOVF import (
    get_decomposition_func,
    get_decomposition_name,
)
from pydec import core
from ..overrides import _auto_registration, _register_builtin_function
from ..exception_utils import none_decomposition_func_error, arg_value_error

from torch.nn.functional import (  # type: ignore[attr-defined]
    _no_grad_embedding_renorm_,
    _get_softmax_dim,
    _mha_shape_check,
    _in_projection_packed,
    _in_projection,
    pad,
    # _scaled_dot_product_attention,
)

# In some cases, these basic types are shadowed by corresponding
# top-level values.  The underscore variants let us refer to these
# types.  See https://github.com/python/mypy/issues/4146 for why these
# workarounds is necessary
# In the future, this will become useful if mypy is introduced into pydec
from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

from typing import (
    List,
    Tuple,
    Optional,
    Union,
    Any,
    ContextManager,
    Callable,
    overload,
    Iterator,
    NamedTuple,
    Sequence,
    TypeVar,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from torch.types import _dtype as DType
else:
    # The JIT doesn't understand Union, nor torch.dtype here
    DType = int

import warnings

T = TypeVar("T")


def _add_docstr(obj: T, doc_obj: str) -> T:
    obj.__doc__ = doc_obj
    return obj


@_auto_registration
def relu(
    input: Composition, inplace: bool = False, *, ref: Optional[Tensor] = None
) -> Composition:
    r"""relu(input, inplace=False, *, ref=None) -> Composition

    Applies the rectified linear unit function element-wise. See
    :class:`~torch.nn.ReLU` for more details.
    """
    if inplace:
        result = pydec.relu_(input, ref=ref)
    else:
        result = pydec.relu(input, ref=ref)
    return result


relu_ = _add_docstr(
    pydec.relu_,
    r"""
relu_(input, *, ref=None) -> Composition

In-place version of :func:`~relu`.
""",
)


@_auto_registration
def leaky_relu(
    input: Tensor,
    negative_slope: float = 0.01,
    inplace: bool = False,
    *,
    ref: Optional[Tensor] = None,
) -> Tensor:
    r"""
    leaky_relu(input, negative_slope=0.01, inplace=False) -> Tensor

    Applies element-wise,
    :math:`\text{LeakyReLU}(x) = \max(0, x) + \text{negative\_slope} * \min(0, x)`

    See :class:`~torch.nn.LeakyReLU` for more details.
    """
    if inplace:
        result = core.decOVF.leaky_relu_(input, negative_slope)
    else:
        result = core.decOVF.leaky_relu(input, negative_slope)
    return result


@_auto_registration
def leaky_relu_(
    input: Composition, negative_slope: float = 0.01, *, ref: Optional[Tensor] = None
) -> Composition:
    return core.decOVF.leaky_relu_(input, negative_slope, ref=ref)


@_auto_registration
def gelu(
    input: Composition, approximate: str = "none", *, ref: Optional[Tensor] = None
) -> Composition:
    return core.decOVF.gelu(input, approximate, ref=ref)


def tanh(input, *, ref: Optional[Tensor] = None):
    r"""tanh(input, *, ref=None) -> Composition

    Applies element-wise,
    :math:`\text{Tanh}(x) = \tanh(x) = \frac{\exp(x) - \exp(-x)}{\exp(x) + \exp(-x)}`

    See :class:`~torch.nn.Tanh` for more details.
    """
    warnings.warn("nn.functional.tanh is deprecated. Use torch.tanh instead.")
    return input.tanh(ref=ref)


def sigmoid(input, *, ref: Optional[Tensor] = None):
    r"""sigmoid(input, *, ref=None) -> Composition

    Applies the element-wise function :math:`\text{Sigmoid}(x) = \frac{1}{1 + \exp(-x)}`

    See :class:`~torch.nn.Sigmoid` for more details.
    """
    warnings.warn("nn.functional.sigmoid is deprecated. Use torch.sigmoid instead.")
    return input.sigmoid(ref=ref)


@_auto_registration
def linear(
    input: Composition, weight: Tensor, bias: Optional[Tensor] = None
) -> Composition:
    out = input @ weight.t()
    if bias is not None:
        out += bias
        return out
    else:
        return out


@_auto_registration
def layer_norm(
    input: Composition,
    normalized_shape: List[int],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    eps: float = 1e-5,
    *,
    ref: Optional[Tensor] = None,
    linearize: bool = True,
) -> Composition:
    r"""Applies Layer Normalization for last certain number of dimensions.

    See :class:`~torch.nn.LayerNorm` for details.
    """
    if linearize:
        return layer_norm_linearized(
            input, normalized_shape, weight, bias, eps, ref=ref
        )
    normalized_dims = tuple(range(-len(normalized_shape), 0))
    input_mean = input.mean(dim=normalized_dims, keepdim=True)
    if ref is None:
        ref = input.c_sum()
    input_std = pydec.sqrt(
        pydec.var(input, dim=normalized_dims, unbiased=False, keepdim=True, ref=ref)
        + eps,
    )
    if weight is not None:
        out = (input - input_mean) * weight / input_std
    else:
        out = (input - input_mean) / input_std

    if bias is not None:
        out += bias
        return out
    else:
        return out


def layer_norm_linearized(
    input: Composition,
    normalized_shape: List[int],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    eps: float = 1e-5,
    *,
    ref: Optional[Tensor] = None,
) -> Composition:
    r"""Applies Layer Normalization for last certain number of dimensions.

    See :class:`~torch.nn.LayerNorm` for details.
    """

    normalized_dims = tuple(range(-len(normalized_shape), 0))
    input_mean = input.mean(dim=normalized_dims, keepdim=True)
    if ref is None:
        ref = input.c_sum()
    input_std = torch.sqrt(
        torch.var(ref, dim=normalized_dims, unbiased=False, keepdim=True) + eps
    )
    if weight is not None:
        out = (input - input_mean) * weight / input_std
    else:
        out = (input - input_mean) / input_std

    if bias is not None:
        out += bias
        return out
    else:
        return out


@overload
def conv2d(
    input: Composition,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[_int, _size] = 1,
    padding: Union[_int, _size] = 0,
    dilation: Union[_int, _size] = 1,
    groups: _int = 1,
) -> Composition:
    ...


@overload
def conv2d(
    input: Composition,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[_int, _size] = 1,
    padding: str = "valid",
    dilation: Union[_int, _size] = 1,
    groups: _int = 1,
) -> Composition:
    ...


@_auto_registration
def conv2d(
    input: Composition,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[_int, _size] = 1,
    padding: Any = 0,
    dilation: Union[_int, _size] = 1,
    groups: _int = 1,
):
    """
    Applies a 2D convolution over an input composition.
    """
    if len(input.size()) != 4 or len(input.size()) != 3:
        raise arg_value_error(
            f"Expected 3D (unbatched) or 4D (batched) input to conv2d, but got input of size: [{len(input.size())}]"
        )
    if len(input.size()) == 3:
        out_component_tensor = F.conv2d(
            input._component_tensor,
            weight,
            None,
            stride,
            padding,
            dilation,
            groups,
        )
        out_residual_tensor = F.conv2d(
            input._residual_tensor,
            weight,
            None,
            stride,
            padding,
            dilation,
            groups,
        )
    else:
        out_component_tensor = F.conv2d(
            input._component_tensor.view((-1,) + input.size()[1:]),
            weight,
            None,
            stride,
            padding,
            dilation,
            groups,
        ).view((-1,) + input.size())
        out_residual_tensor = F.conv2d(
            input._residual_tensor,
            weight,
            None,
            stride,
            padding,
            dilation,
            groups,
        )
    out_residual_tensor += bias
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


@_auto_registration
def max_pool2d(
    input: Composition,
    kernel_size: Union[_int, Tuple[_int, _int]],
    stride: Union[_int, Tuple[_int, _int]] = None,
    padding: Union[_int, Tuple[_int, _int]] = 0,
    dilation: Union[_int, Tuple[_int, _int]] = 1,
    ceil_mode: _bool = False,
    return_indices: _bool = False,
):
    """
    Applies a 2D max pooling over an input composition.
    """
    if len(input.size()) != 4 or len(input.size()) != 3:
        raise arg_value_error(
            f"Expected 3D (unbatched) or 4D (batched) input to conv2d, but got input of size: [{len(input.size())}]"
        )
    recovery = input.c_sum()
    _, indices = F.max_pool2d(
        recovery,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        ceil_mode=ceil_mode,
        return_indices=True,
    )

    out = max_pool2d_with_indices(input, indices=indices)
    if return_indices:
        return out, indices
    else:
        return out


def max_pool2d_with_indices(
    input: Composition,
    indices: Tensor,
    kernel_size: Union[_int, Tuple[_int, _int]],
    stride: Union[_int, Tuple[_int, _int]] = None,
    padding: Union[_int, Tuple[_int, _int]] = 0,
    dilation: Union[_int, Tuple[_int, _int]] = 1,
    ceil_mode: _bool = False,
):
    """
    Applies a 2D max pooling over an input composition.
    TODO: need to support arguments [kernel_size, stride, padding, dilation, ceil_mode]
    """
    # TODO: api name is duplicated with torch.nn.functional.max_pool2d_with_indices
    if len(input.size()) != 4 or len(input.size()) != 3:
        raise arg_value_error(
            f"Expected 3D (unbatched) or 4D (batched) input to conv2d, but got input of size: [{len(input.size())}]"
        )
    flat_input = input.view(input.size()[:-2] + (-1,))
    H_out, W_out = indices.size()[-2:]

    out = flat_input.gather(dim=-1, index=indices.view(indices.size()[:-2] + (-1,)))
    out = out.view(out.size()[:-1] + (H_out, W_out))

    return out


@_auto_registration
def dropout(
    input: Composition, p: float = 0.5, training: bool = True, inplace: bool = False
) -> Tensor:
    r"""
    During training, randomly zeroes some of the elements of the input
    tensor with probability :attr:`p` using samples from a Bernoulli
    distribution.

    See :class:`~torch.nn.Dropout` for details.

    Args:
        p: probability of an element to be zeroed. Default: 0.5
        training: apply dropout if is ``True``. Default: ``True``
        inplace: If set to ``True``, will do this operation in-place. Default: ``False``
    """
    if p < 0.0 or p > 1.0:
        raise ValueError(
            "dropout probability has to be between 0 and 1, " "but got {}".format(p)
        )
    if not training:
        return input

    drop_mask = torch.ones(input.size(), device=input.devce, dtype=input.dtype)
    drop_mask = torch._VF.dropout_(drop_mask, p, training)
    if inplace:
        input *= drop_mask
        return input
    else:
        return input * drop_mask


@_auto_registration
def embedding(
    input: IndexComposition,
    weight: Tensor,
    padding_idx: Optional[int] = None,
    max_norm: Optional[float] = None,
    norm_type: float = 2.0,
    scale_grad_by_freq: bool = False,
    sparse: bool = False,
) -> Composition:
    r"""A simple lookup table that looks up embeddings in a fixed dictionary and size.

    This module is often used to retrieve word embeddings using indices.
    The input to the module is a list of indices, and the embedding matrix,
    and the output is the corresponding word embeddings.

    See :class:`torch.nn.Embedding` for more details.

    Args:
        input (IndexComposition): IndexComposition containing indices into the embedding matrix
        weight (Tensor): The embedding matrix with number of rows equal to the maximum possible index + 1,
            and number of columns equal to the embedding size
        padding_idx (int, optional): If specified, the entries at :attr:`padding_idx` do not contribute to the gradient;
                                     therefore, the embedding vector at :attr:`padding_idx` is not updated during training,
                                     i.e. it remains as a fixed "pad".
        max_norm (float, optional): If given, each embedding vector with norm larger than :attr:`max_norm`
                                    is renormalized to have norm :attr:`max_norm`.
                                    Note: this will modify :attr:`weight` in-place.
        norm_type (float, optional): The p of the p-norm to compute for the :attr:`max_norm` option. Default ``2``.
        scale_grad_by_freq (boolean, optional): If given, this will scale gradients by the inverse of frequency of
                                                the words in the mini-batch. Default ``False``.
        sparse (bool, optional): If ``True``, gradient w.r.t. :attr:`weight` will be a sparse tensor. See Notes under
                                 :class:`torch.nn.Embedding` for more details regarding sparse gradients.

    Shape:
        - Input: LongTensor of arbitrary shape containing the indices to extract
        - Weight: Embedding matrix of floating point type with shape `(V, embedding_dim)`,
          where V = maximum index + 1 and embedding_dim = the embedding size
        - Output: `(*, embedding_dim)`, where `*` is the input shape

    Examples::

        >>> # a batch of 2 samples of 4 indices each
        >>> input = torch.tensor([[1,2,4,5],[4,3,2,9]])
        >>> # an embedding matrix containing 10 tensors of size 3
        >>> embedding_matrix = torch.rand(10, 3)
        >>> F.embedding(input, embedding_matrix)
        tensor([[[ 0.8490,  0.9625,  0.6753],
                 [ 0.9666,  0.7761,  0.6108],
                 [ 0.6246,  0.9751,  0.3618],
                 [ 0.4161,  0.2419,  0.7383]],

                [[ 0.6246,  0.9751,  0.3618],
                 [ 0.0237,  0.7794,  0.0528],
                 [ 0.9666,  0.7761,  0.6108],
                 [ 0.3385,  0.8612,  0.1867]]])

        >>> # example with padding_idx
        >>> weights = torch.rand(10, 3)
        >>> weights[0, :].zero_()
        >>> embedding_matrix = weights
        >>> input = torch.tensor([[0,2,0,5]])
        >>> F.embedding(input, embedding_matrix, padding_idx=0)
        tensor([[[ 0.0000,  0.0000,  0.0000],
                 [ 0.5609,  0.5384,  0.8720],
                 [ 0.0000,  0.0000,  0.0000],
                 [ 0.6262,  0.2438,  0.7471]]])
    """
    if padding_idx is not None:
        if padding_idx > 0:
            assert padding_idx < weight.size(
                0
            ), "Padding_idx must be within num_embeddings"
        elif padding_idx < 0:
            assert padding_idx >= -weight.size(
                0
            ), "Padding_idx must be within num_embeddings"
            padding_idx = weight.size(0) + padding_idx
    else:
        padding_idx = -1
    if max_norm is not None:
        # Note [embedding_renorm contiguous]
        # `embedding_renorm_` will call .contiguous() on input anyways, so we
        # call it here and take advantage of the improved locality in the
        # `embedding` call below too.
        input = input.contiguous()
        # Note [embedding_renorm set_grad_enabled]
        # XXX: equivalent to
        # with torch.no_grad():
        #   torch.embedding_renorm_
        # remove once script supports set_grad_enabled
        _no_grad_embedding_renorm_(weight, input, max_norm, norm_type)

    component_empty_mask, residual_empty_mask = input.empty_mask
    component_tensor = input._component_tensor.masked_fill(
        component_empty_mask, padding_idx
    )
    residual_tensor = input._residual_tensor.masked_fill(
        residual_empty_mask, padding_idx
    )
    out_component_tensor = torch.embedding(
        weight, component_tensor, padding_idx, scale_grad_by_freq, sparse
    )
    out_residual_tensor = torch.embedding(
        weight, residual_tensor, padding_idx, scale_grad_by_freq, sparse
    )
    out_component_tensor[component_empty_mask] = 0
    out_residual_tensor[residual_empty_mask] = 0
    return pydec.as_composition(out_component_tensor, out_residual_tensor)


def legacy_relu(input: Composition, ref: Optional[Tensor] = None) -> Composition:
    # TODO: maybe deprecated after we have scaling decomposition
    if ref is None:
        ref = input.c_sum()
    zero_mask = ref < 0
    residual_out = torch.nn.functional.relu(input._residual_tensor)
    out = input.masked_fill(zero_mask, 0.0)
    masked_residual_out = out._residual_tensor
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        delta_context = pydec.as_composition(input._component_tensor, residual_out)
        delta_out = decomposition_func(
            input=delta_context, func=lambda x: x, ref=masked_residual_out
        )
        out._component_tensor += delta_out._component_tensor
        out._residual_tensor = delta_out._residual_tensor
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


@_auto_registration
def softmax(
    input: Composition,
    dim: Optional[int] = None,
    _stacklevel: int = 3,
    dtype: Optional[DType] = None,
) -> Composition:
    r"""Applies a softmax function.

    Softmax is defined as:

    :math:`\text{Softmax}(x_{i}) = \frac{\exp(x_i)}{\sum_j \exp(x_j)}`

    It is applied to all slices along dim, and will re-scale them so that the elements
    lie in the range `[0, 1]` and sum to 1.

    See :class:`~torch.nn.Softmax` for more details.

    Args:
        input (Composition): input
        dim (int): A dimension along which softmax will be computed.
        dtype (:class:`torch.dtype`, optional): the desired data type of returned composition.
          If specified, the input composition is casted to :attr:`dtype` before the operation
          is performed. This is useful for preventing data type overflows. Default: None.

    .. note::
        This function doesn't work directly with NLLLoss,
        which expects the Log to be computed between the Softmax and itself.
        Use log_softmax instead (it's faster and has better numerical properties).

    """
    if dim is None:
        dim = _get_softmax_dim("softmax", input.dim(), _stacklevel)
    if dtype is None:
        ret = input.softmax(dim)
    else:
        ret = input.softmax(dim, dtype=dtype)
    return ret


@_auto_registration
def multi_head_attention_forward(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    embed_dim_to_check: int,
    num_heads: int,
    in_proj_weight: Optional[Tensor],
    in_proj_bias: Optional[Tensor],
    bias_k: Optional[Tensor],
    bias_v: Optional[Tensor],
    add_zero_attn: bool,
    dropout_p: float,
    out_proj_weight: Tensor,
    out_proj_bias: Optional[Tensor],
    training: bool = True,
    key_padding_mask: Optional[Tensor] = None,
    need_weights: bool = True,
    attn_mask: Optional[Tensor] = None,
    use_separate_proj_weight: bool = False,
    q_proj_weight: Optional[Tensor] = None,
    k_proj_weight: Optional[Tensor] = None,
    v_proj_weight: Optional[Tensor] = None,
    static_k: Optional[Tensor] = None,
    static_v: Optional[Tensor] = None,
    average_attn_weights: bool = True,
) -> Tuple[Tensor, Optional[Tensor]]:
    r"""
    Args:
        query, key, value: map a query and a set of key-value pairs to an output.
            See "Attention Is All You Need" for more details.
        embed_dim_to_check: total dimension of the model.
        num_heads: parallel attention heads.
        in_proj_weight, in_proj_bias: input projection weight and bias.
        bias_k, bias_v: bias of the key and value sequences to be added at dim=0.
        add_zero_attn: add a new batch of zeros to the key and
                       value sequences at dim=1.
        dropout_p: probability of an element to be zeroed.
        out_proj_weight, out_proj_bias: the output projection weight and bias.
        training: apply dropout if is ``True``.
        key_padding_mask: if provided, specified padding elements in the key will
            be ignored by the attention. This is an binary mask. When the value is True,
            the corresponding value on the attention layer will be filled with -inf.
        need_weights: output attn_output_weights.
        attn_mask: 2D or 3D mask that prevents attention to certain positions. A 2D mask will be broadcasted for all
            the batches while a 3D mask allows to specify a different mask for the entries of each batch.
        use_separate_proj_weight: the function accept the proj. weights for query, key,
            and value in different forms. If false, in_proj_weight will be used, which is
            a combination of q_proj_weight, k_proj_weight, v_proj_weight.
        q_proj_weight, k_proj_weight, v_proj_weight, in_proj_bias: input projection weight and bias.
        static_k, static_v: static key and value used for attention operators.
        average_attn_weights: If true, indicates that the returned ``attn_weights`` should be averaged across heads.
            Otherwise, ``attn_weights`` are provided separately per head. Note that this flag only has an effect
            when ``need_weights=True.``. Default: True


    Shape:
        Inputs:
        - query: :math:`(L, E)` or :math:`(L, N, E)` where L is the target sequence length, N is the batch size, E is
          the embedding dimension.
        - key: :math:`(S, E)` or :math:`(S, N, E)`, where S is the source sequence length, N is the batch size, E is
          the embedding dimension.
        - value: :math:`(S, E)` or :math:`(S, N, E)` where S is the source sequence length, N is the batch size, E is
          the embedding dimension.
        - key_padding_mask: :math:`(S)` or :math:`(N, S)` where N is the batch size, S is the source sequence length.
          If a ByteTensor is provided, the non-zero positions will be ignored while the zero positions
          will be unchanged. If a BoolTensor is provided, the positions with the
          value of ``True`` will be ignored while the position with the value of ``False`` will be unchanged.
        - attn_mask: 2D mask :math:`(L, S)` where L is the target sequence length, S is the source sequence length.
          3D mask :math:`(N*num_heads, L, S)` where N is the batch size, L is the target sequence length,
          S is the source sequence length. attn_mask ensures that position i is allowed to attend the unmasked
          positions. If a ByteTensor is provided, the non-zero positions are not allowed to attend
          while the zero positions will be unchanged. If a BoolTensor is provided, positions with ``True``
          are not allowed to attend while ``False`` values will be unchanged. If a FloatTensor
          is provided, it will be added to the attention weight.
        - static_k: :math:`(N*num_heads, S, E/num_heads)`, where S is the source sequence length,
          N is the batch size, E is the embedding dimension. E/num_heads is the head dimension.
        - static_v: :math:`(N*num_heads, S, E/num_heads)`, where S is the source sequence length,
          N is the batch size, E is the embedding dimension. E/num_heads is the head dimension.

        Outputs:
        - attn_output: :math:`(L, E)` or :math:`(L, N, E)` where L is the target sequence length, N is the batch size,
          E is the embedding dimension.
        - attn_output_weights: Only returned when ``need_weights=True``. If ``average_attn_weights=True``, returns
          attention weights averaged across heads of shape :math:`(L, S)` when input is unbatched or
          :math:`(N, L, S)`, where :math:`N` is the batch size, :math:`L` is the target sequence length, and
          :math:`S` is the source sequence length. If ``average_weights=False``, returns attention weights per
          head of shape :math:`(num_heads, L, S)` when input is unbatched or :math:`(N, num_heads, L, S)`.
    """
    return NotImplemented

    is_batched = _mha_shape_check(
        query, key, value, key_padding_mask, attn_mask, num_heads
    )

    # For unbatched input, we unsqueeze at the expected batch-dim to pretend that the input
    # is batched, run the computation and before returning squeeze the
    # batch dimension so that the output doesn't carry this temporary batch dimension.
    if not is_batched:
        # unsqueeze if the input is unbatched
        query = query.unsqueeze(1)
        key = key.unsqueeze(1)
        value = value.unsqueeze(1)
        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.unsqueeze(0)

    # set up shape vars
    tgt_len, bsz, embed_dim = query.shape
    src_len, _, _ = key.shape
    assert (
        embed_dim == embed_dim_to_check
    ), f"was expecting embedding dimension of {embed_dim_to_check}, but got {embed_dim}"
    if isinstance(embed_dim, torch.Tensor):
        # embed_dim can be a tensor when JIT tracing
        head_dim = embed_dim.div(num_heads, rounding_mode="trunc")
    else:
        head_dim = embed_dim // num_heads
    assert (
        head_dim * num_heads == embed_dim
    ), f"embed_dim {embed_dim} not divisible by num_heads {num_heads}"
    if use_separate_proj_weight:
        # allow MHA to have different embedding dimensions when separate projection weights are used
        assert (
            key.shape[:2] == value.shape[:2]
        ), f"key's sequence and batch dims {key.shape[:2]} do not match value's {value.shape[:2]}"
    else:
        assert (
            key.shape == value.shape
        ), f"key shape {key.shape} does not match value shape {value.shape}"

    #
    # compute in-projection
    #
    if not use_separate_proj_weight:
        assert (
            in_proj_weight is not None
        ), "use_separate_proj_weight is False but in_proj_weight is None"
        q, k, v = _in_projection_packed(query, key, value, in_proj_weight, in_proj_bias)
    else:
        assert (
            q_proj_weight is not None
        ), "use_separate_proj_weight is True but q_proj_weight is None"
        assert (
            k_proj_weight is not None
        ), "use_separate_proj_weight is True but k_proj_weight is None"
        assert (
            v_proj_weight is not None
        ), "use_separate_proj_weight is True but v_proj_weight is None"
        if in_proj_bias is None:
            b_q = b_k = b_v = None
        else:
            b_q, b_k, b_v = in_proj_bias.chunk(3)
        q, k, v = _in_projection(
            query,
            key,
            value,
            q_proj_weight,
            k_proj_weight,
            v_proj_weight,
            b_q,
            b_k,
            b_v,
        )

    # prep attention mask
    if attn_mask is not None:
        if attn_mask.dtype == torch.uint8:
            warnings.warn(
                "Byte tensor for attn_mask in nn.MultiheadAttention is deprecated. Use bool tensor instead."
            )
            attn_mask = attn_mask.to(torch.bool)
        else:
            assert (
                attn_mask.is_floating_point() or attn_mask.dtype == torch.bool
            ), f"Only float, byte, and bool types are supported for attn_mask, not {attn_mask.dtype}"
        # ensure attn_mask's dim is 3
        if attn_mask.dim() == 2:
            correct_2d_size = (tgt_len, src_len)
            if attn_mask.shape != correct_2d_size:
                raise RuntimeError(
                    f"The shape of the 2D attn_mask is {attn_mask.shape}, but should be {correct_2d_size}."
                )
            attn_mask = attn_mask.unsqueeze(0)
        elif attn_mask.dim() == 3:
            correct_3d_size = (bsz * num_heads, tgt_len, src_len)
            if attn_mask.shape != correct_3d_size:
                raise RuntimeError(
                    f"The shape of the 3D attn_mask is {attn_mask.shape}, but should be {correct_3d_size}."
                )
        else:
            raise RuntimeError(
                f"attn_mask's dimension {attn_mask.dim()} is not supported"
            )

    # prep key padding mask
    if key_padding_mask is not None and key_padding_mask.dtype == torch.uint8:
        warnings.warn(
            "Byte tensor for key_padding_mask in nn.MultiheadAttention is deprecated. Use bool tensor instead."
        )
        key_padding_mask = key_padding_mask.to(torch.bool)

    # add bias along batch dimension (currently second)
    if bias_k is not None and bias_v is not None:
        assert static_k is None, "bias cannot be added to static key."
        assert static_v is None, "bias cannot be added to static value."
        k = torch.cat([k, bias_k.repeat(1, bsz, 1)])
        v = torch.cat([v, bias_v.repeat(1, bsz, 1)])
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))
    else:
        assert bias_k is None
        assert bias_v is None

    #
    # reshape q, k, v for multihead attention and make em batch first
    #
    q = q.contiguous().view(tgt_len, bsz * num_heads, head_dim).transpose(0, 1)
    if static_k is None:
        k = k.contiguous().view(k.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        # TODO finish disentangling control flow so we don't do in-projections when statics are passed
        assert (
            static_k.size(0) == bsz * num_heads
        ), f"expecting static_k.size(0) of {bsz * num_heads}, but got {static_k.size(0)}"
        assert (
            static_k.size(2) == head_dim
        ), f"expecting static_k.size(2) of {head_dim}, but got {static_k.size(2)}"
        k = static_k
    if static_v is None:
        v = v.contiguous().view(v.shape[0], bsz * num_heads, head_dim).transpose(0, 1)
    else:
        # TODO finish disentangling control flow so we don't do in-projections when statics are passed
        assert (
            static_v.size(0) == bsz * num_heads
        ), f"expecting static_v.size(0) of {bsz * num_heads}, but got {static_v.size(0)}"
        assert (
            static_v.size(2) == head_dim
        ), f"expecting static_v.size(2) of {head_dim}, but got {static_v.size(2)}"
        v = static_v

    # add zero attention along batch dimension (now first)
    if add_zero_attn:
        zero_attn_shape = (bsz * num_heads, 1, head_dim)
        k = torch.cat(
            [k, torch.zeros(zero_attn_shape, dtype=k.dtype, device=k.device)], dim=1
        )
        v = torch.cat(
            [v, torch.zeros(zero_attn_shape, dtype=v.dtype, device=v.device)], dim=1
        )
        if attn_mask is not None:
            attn_mask = pad(attn_mask, (0, 1))
        if key_padding_mask is not None:
            key_padding_mask = pad(key_padding_mask, (0, 1))

    # update source sequence length after adjustments
    src_len = k.size(1)

    # merge key padding and attention masks
    if key_padding_mask is not None:
        assert key_padding_mask.shape == (
            bsz,
            src_len,
        ), f"expecting key_padding_mask shape of {(bsz, src_len)}, but got {key_padding_mask.shape}"
        key_padding_mask = (
            key_padding_mask.view(bsz, 1, 1, src_len)
            .expand(-1, num_heads, -1, -1)
            .reshape(bsz * num_heads, 1, src_len)
        )
        if attn_mask is None:
            attn_mask = key_padding_mask
        elif attn_mask.dtype == torch.bool:
            attn_mask = attn_mask.logical_or(key_padding_mask)
        else:
            attn_mask = attn_mask.masked_fill(key_padding_mask, float("-inf"))

    # convert mask to float
    if attn_mask is not None and attn_mask.dtype == torch.bool:
        new_attn_mask = torch.zeros_like(attn_mask, dtype=q.dtype)
        new_attn_mask.masked_fill_(attn_mask, float("-inf"))
        attn_mask = new_attn_mask

    # adjust dropout probability
    if not training:
        dropout_p = 0.0

    #
    # (deep breath) calculate attention and out projection
    #
    attn_output, attn_output_weights = _scaled_dot_product_attention(
        q, k, v, attn_mask, dropout_p
    )
    attn_output = (
        attn_output.transpose(0, 1).contiguous().view(tgt_len * bsz, embed_dim)
    )
    attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
    attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))

    if need_weights:
        # optionally average attention weights over heads
        attn_output_weights = attn_output_weights.view(bsz, num_heads, tgt_len, src_len)
        if average_attn_weights:
            attn_output_weights = attn_output_weights.sum(dim=1) / num_heads

        if not is_batched:
            # squeeze the output if input was unbatched
            attn_output = attn_output.squeeze(1)
            attn_output_weights = attn_output_weights.squeeze(0)
        return attn_output, attn_output_weights
    else:
        if not is_batched:
            # squeeze the output if input was unbatched
            attn_output = attn_output.squeeze(1)
        return attn_output, None



import torch.nn as nn
from pydec import Composition
from .. import functional as F
from .module import DecModule
from .meta import ProxyModule



class ReLU(nn.ReLU, DecModule, metaclass=ProxyModule):
    def pydec_forward(self, input: Composition) -> Composition:
            return F.relu(input, inplace=self.inplace)


from .module import DecModule, Module
from .linear import Identity, Linear
from .container import (
    Container,
    Sequential,
    ModuleList,
    ModuleDict,
    ParameterList,
    ParameterDict,
)
from .activation import ReLU
from . import meta



"""
Decomposition of One Variable Functions.
"""

from __future__ import annotations
import torch
import pydec
from torch import Tensor
import functools

from typing import Dict, Tuple, Union, Any, Callable, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..._composition import Composition


from .states import (
    get_decomposition_func,
    get_decomposition_name,
)

from pydec.exception_utils import none_decomposition_func_error

__all__ = [
    "relu",
    "relu_",
    "leaky_relu",
    "leaky_relu_",
    "gelu",
    "tanh",
    "tanh_",
    "sigmoid",
    "sigmoid_",
    "reciprocal",
    "reciprocal_",
    "exp",
    "exp_",
    "biased_exp",
    "sqrt",
    "sqrt_",
    "square",
    "square_",
]


def relu(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    # TODO: maybe optimize the relu in scaling algorithm for speed .
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        # TODO: inplace arg overwrite
        out = decomposition_func(input=input, func=torch.nn.functional.relu, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def relu_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        # TODO: inplace arg overwrite
        out = decomposition_func(
            input=input, func=torch.nn.functional.relu_, inplace=True, ref=ref
        )
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def leaky_relu(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(
            input=input, func=torch.nn.functional.leaky_relu, ref=ref
        )
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def leaky_relu_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(
            input=input, func=torch.nn.functional.leaky_relu, ref=ref, inplace=True
        )
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def gelu(
    input: Composition, approximate: str = "none", *, ref: Optional[Tensor] = None
) -> Composition:
    torch_gelu = functools.partial(torch.nn.functional.gelu, approximate=approximate)
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch_gelu, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def tanh(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.tanh, out=out, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def tanh_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.tanh_, inplace=True, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def sigmoid(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.sigmoid, out=out, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def sigmoid_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(
            input=input, func=torch.sigmoid_, inplace=True, ref=ref
        )
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def reciprocal(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.reciprocal, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def reciprocal_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(
            input=input, func=torch.reciprocal_, inplace=True, ref=ref
        )
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def exp(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    # TODO: need a special dec to address numerical  explosion
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.exp, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def exp_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.exp_, inplace=True, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def biased_exp(
    input: Composition,
    bias: Tensor,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    # TODO: need a special dec to address numerical  explosion
    if ref is None:
        ref = input.c_sum()
    new_input = pydec.as_composition(input.components, input.residual + bias)
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=new_input, func=torch.exp, ref=ref + bias)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


# def biased_exp_stable(
#     input: Composition,
#     bias: Tensor,
#     *,
#     out: Optional[Composition] = None,
#     ref: Optional[Tensor] = None,
# ) -> Composition:
#     if ref is None:
#         ref = input.c_sum()
#     new_components = torch.cat([input.components, input.residual[None]])
#     new_residual = torch.zeros_like(input.residual).fill_(-torch.inf)
#     new_input = pydec.as_composition(new_components, new_residual)
#     decomposition_func = get_decomposition_func()
#     if decomposition_func is not None:
#         out = decomposition_func(input=new_input, func=torch.exp, ref=ref + bias)
#         assert isinstance(out, pydec.Composition)
#         out_components = out.components[:-1]
#         out_residual = out.components[-1]
#         return pydec.as_composition(out_components, out_residual)
#     else:
#         raise none_decomposition_func_error(get_decomposition_name())


def sqrt(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.sqrt, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def sqrt_(input: Composition, *, ref: Optional[Tensor] = None) -> Composition:
    decomposition_func = get_decomposition_func()
    if decomposition_func is not None:
        out = decomposition_func(input=input, func=torch.sqrt_, ref=ref)
        assert isinstance(out, pydec.Composition)
        return out
    else:
        raise none_decomposition_func_error(get_decomposition_name())


def square(
    input: Composition,
    *,
    out: Optional[Composition] = None,
    ref: Optional[Tensor] = None,
) -> Composition:
    if ref is None:
        ref = input.c_sum()
    out_residual = torch.square(
        input.residual, out=out.residual if out is not None else None
    )
    out_components = torch.mul(
        input.components,
        ref + input.residual,
        out=out.components if out is not None else None,
    )
    return pydec.as_composition(out_components, out_residual)


def square_(
    input: Composition,
    ref: Optional[Tensor] = None,
) -> Composition:
    if ref is None:
        ref = input.c_sum()
    ref = ref + input.residual
    input.residual.square_()
    input.components.mul_(ref)
    return input



from __future__ import annotations

import torch
from . import core

from typing import Any, Dict, Union, List, Tuple, Sequence, Optional, Callable, overload
from torch import Tensor
from torch._C import memory_format
from torch.autograd.grad_mode import _DecoratorContextManager
import pydec
from ._composition_str import _c_str
import types
import sys
from .overrides import (
    _register_builtin_function,
    _auto_registration,
    is_registered,
    dispatch_torch_function,
)

# In some cases, these basic types are shadowed by corresponding
# top-level values.  The underscore variants let us refer to these
# types.  See https://github.com/python/mypy/issues/4146 for why these
# workarounds is necessary
# In the future, this will become useful if mypy is introduced into pydec
from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)

from pydec.exception_utils import (
    args_error,
    size_error,
    component_num_error,
    unsupported_operand_error,
    arg_value_error,
    none_decomposition_func_error,
)


from pydec.utils import _shift_dim, _shift_dims, parse_args

r"""
To avoid circular import, we have to initialize the following method in __init__.py.
"""


class Composition:
    r"""
    TODO: Composition doc
    """

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if not is_registered(func) or not all(
            issubclass(t, (torch.Tensor, pydec.Composition)) for t in types
        ):
            return NotImplemented
        return dispatch_torch_function(func)(*args, **kwargs)

    @property
    def requires_grad(self) -> _bool:
        return self._residual_tensor.requires_grad

    @property
    def shape(self) -> torch.Size:
        return self._residual_tensor.shape

    @property
    def device(self) -> _device:
        return self._residual_tensor.device

    @property
    def dtype(self) -> _dtype:
        return self._residual_tensor.dtype

    @property
    def layout(self) -> _dtype:
        return self._residual_tensor.layout

    @property
    def T(self) -> Composition:
        return self.permute(*torch.arange(self.ndim - 1, -1, -1))

    @property
    def mT(self) -> Composition:
        return self.transpose(-2, -1)

    @property
    def ndim(self) -> _int:
        return self.dim()

    @property
    def components(self) -> Tensor:
        return self._component_tensor

    @property
    def residual(self) -> Tensor:
        return self._residual_tensor

    @property
    def recovery(self) -> Tensor:
        return self.c_sum()

    @property
    def is_cuda(self):
        return self._residual_tensor.is_cuda

    @overload
    def __init__(
        self, component_tensor: Tensor, residual_tensor: Optional[Tensor] = None
    ) -> None:
        ...

    @overload
    def __init__(self, composition: Composition) -> None:
        ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        def init_from_tensor(
            component_tensor: Tensor, residual_tensor: Optional[Tensor] = None
        ):
            if residual_tensor is not None:
                if component_tensor.dtype != residual_tensor.dtype:
                    raise arg_value_error(
                        f"the dtype of component_tensor ({component_tensor.dtype}) should be the same as that of residual_tensor ({residual_tensor.dtype})"
                    )
                if component_tensor.device != residual_tensor.device:
                    raise arg_value_error(
                        f"the device of component_tensor ({component_tensor.device}) should be the same as that of residual_tensor ({residual_tensor.device})"
                    )
                if component_tensor.requires_grad != residual_tensor.requires_grad:
                    raise arg_value_error(
                        f"component_tensor.requires_grad ({component_tensor.requires_grad}) should be the same as residual_tensor.requires_grad ({residual_tensor.requires_grad})"
                    )

            # formal way but may cause problems, see https://github.com/pytorch/pytorch/issues/85094
            self._component_tensor = component_tensor.clone().detach()
            if residual_tensor is not None:
                if component_tensor.size()[1:] != residual_tensor.size():
                    raise size_error(
                        component_tensor.size()[1:],
                        residual_tensor.size(),
                        "composition",
                        "residual",
                    )
                self._residual_tensor = residual_tensor.clone().detach()
            else:
                self._residual_tensor = torch.zeros(component_tensor.size()[1:]).to(
                    component_tensor
                )

        self._component_tensor: Tensor = None  # type: ignore[no-redef, assignment]
        self._residual_tensor: Tensor = None  # type: ignore[no-redef, assignment]
        input_kwargs = kwargs.copy()  # for error hint

        if len(args) == 0 and len(kwargs) == 0:
            # a private constructor to create a void composition with no data,
            # which is usually assigned in `_from_replace`.
            return
        elif len(args) > 0:
            if isinstance(args[0], Composition):
                parse_args(args, ["composition"], kwargs)
            else:
                parse_args(args, ["component_tensor", "residual_tensor"], kwargs)
        elif len(args) > 2:
            raise args_error("Composition.__init__", args, input_kwargs)

        if "composition" in kwargs:
            c = kwargs["composition"]
            if isinstance(c, Composition):
                init_from_tensor(c._component_tensor, c._residual_tensor)
            else:
                raise args_error("Composition.__init__", args, input_kwargs)
        elif "component_tensor" in kwargs:
            if "residual_tensor" in kwargs and not isinstance(
                kwargs["residual_tensor"], Tensor
            ):
                raise args_error("Composition.__init__", args, input_kwargs)
            if isinstance(kwargs["component_tensor"], Tensor):
                init_from_tensor(**kwargs)
            else:
                raise args_error("Composition.__init__", args, input_kwargs)
        else:
            raise args_error("Composition.__init__", args, input_kwargs)

    @_auto_registration
    def __getitem__(
        self, indices: Union[None, _int, slice, Tensor, List, Tuple]
    ) -> Composition:
        if not pydec.is_c_accessing_enabled():
            if isinstance(indices, tuple):
                indices = (slice(None, None, None),) + indices
            else:
                indices = (
                    slice(None, None, None),
                    indices,
                )
        return self.__c_getitem__(indices)  # type: ignore[return-value]

    @_auto_registration
    def __setitem__(
        self,
        indices: Union[None, _int, slice, Tensor, List, Tuple],
        val: Union[Composition, Tensor, Number],
    ) -> None:
        if not pydec.is_c_accessing_enabled():
            if isinstance(indices, tuple):
                indices = (slice(None, None, None),) + indices
            else:
                indices = (
                    slice(None, None, None),
                    indices,
                )
        return self.__c_setitem__(indices, val)

    def __c_getitem__(
        self, indices: Union[None, _int, slice, Tensor, List, Tuple]
    ) -> Union[Composition, Tensor]:
        if isinstance(indices, (type(None), _int, slice, List, Tensor)):
            indices = (indices,)
        assert isinstance(indices, tuple)  # make mypy ignore [index] errors
        if indices[0] is None:
            raise arg_value_error(
                "The first dimension of indices should not be NoneType"
            )
        if isinstance(indices[0], _int):
            return self._component_tensor[indices]
        else:
            out_component_tensor = self._component_tensor[indices]
            out_residual_tensor = self._residual_tensor[indices[1:]]
            return pydec.as_composition(out_component_tensor, out_residual_tensor)

    def __c_setitem__(
        self,
        indices: Union[None, _int, slice, Tensor, List, Tuple],
        val: Union[Composition, Tensor, Number],
    ) -> None:
        if isinstance(indices, (type(None), _int, slice, List, Tensor)):
            indices = (indices,)
        assert isinstance(indices, tuple)  # make mypy ignore [index] errors
        if indices[0] is None:
            raise arg_value_error(
                "The first dimension of indices should not be NoneType"
            )

        if isinstance(val, (Tensor, _int, _float, _bool)):
            # TODO: also set item for residual? `c[:]=0` not work with c.residual
            self._component_tensor[indices] = val
            return
        if isinstance(val, (Composition)):
            if isinstance(indices[0], _int):
                raise arg_value_error(
                    f"Expected the assignment value to be (Tensor) or (Number), not ({type(val).__name__}) for single component assignment"
                )
            else:
                self._component_tensor[indices] = val._component_tensor
                self._residual_tensor[indices[1:]] = val._residual_tensor

    def __call__(self) -> _C_AccessingComposition:
        """
        Shortcut for `__c_getitem__` and `__c_setitem__`, e.g., `c()[2]` equals to `c.__c_getitem__(2)`.
        """
        return _C_AccessingComposition(self)

    @_auto_registration
    def __len__(self):
        if pydec.is_c_accessing_enabled():
            return self.components.__len__()
        else:
            return self.residual.__len__()

    def __iter__(self):
        if pydec.is_c_accessing_enabled():
            # TODO: should return composition
            return self._component_tensor.__iter__()
        else:
            # TODO: return a iterator
            raise NotImplementedError

    @_auto_registration
    def __reversed__(self):
        return self._component_tensor.__reversed__()

    @_auto_registration
    def __contains__(self, element):
        return self._component_tensor.__contains__(element)

    @_auto_registration  # type: ignore[arg-type]
    def __repr__(
        self, *, composition_contents: Optional[List[Union[str, None]]] = None
    ) -> str:
        return _c_str(self, composition_contents=composition_contents)

    @_auto_registration
    def numel(self) -> _int:
        return pydec.numel(self)

    def c_numel(self, count_residual=False) -> _int:
        return pydec.c_numel(self, count_residual=count_residual)

    def numc(self) -> _int:
        return pydec.numc(self)

    @_auto_registration
    def clone(self, *, memory_format: Optional[memory_format] = None) -> Composition:
        return pydec.clone(self, memory_format=memory_format)

    @_auto_registration
    def detach(self) -> Composition:
        return pydec.detach(self)

    @_auto_registration
    def detach_(self) -> Composition:
        return pydec.detach_(self)

    @overload
    def size(self) -> torch.Size:
        ...

    @overload
    def size(self, dim: _int) -> _int:
        ...

    @overload
    def size(self, dim: Optional[_int] = None) -> Union[torch.Size, _int]:
        ...

    @_auto_registration
    def size(self, dim: Optional[_int] = None) -> Union[torch.Size, _int]:
        if dim is None:
            return self.c_size()[1:]
        else:
            return self.size()[dim]

    @overload
    def c_size(self) -> torch.Size:
        ...

    @overload
    def c_size(self, dim: _int) -> _int:
        ...

    def c_size(self, dim: Optional[_int] = None) -> Union[torch.Size, _int]:
        if dim is None:
            return self._component_tensor.size()
        else:
            return self._component_tensor.size(dim)

    @_auto_registration
    def dim(self) -> _int:
        return self._residual_tensor.dim()

    @_auto_registration
    def __neg__(self) -> Composition:
        return pydec.as_composition(-self._component_tensor, -self._residual_tensor)

    @_auto_registration
    def __pos__(self) -> Composition:
        return pydec.as_composition(+self._component_tensor, +self._residual_tensor)

    @_auto_registration
    def __iadd__(self, other) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("+=", type(self), type(other))
        if isinstance(other, Composition):
            return core.decBLAS.cc_add_(self, other)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_add_(self, other)
        else:
            raise unsupported_operand_error("+=", type(self), type(other))

    @_auto_registration
    def __add__(self, other) -> Composition:
        try:
            return pydec.add(self, other)
        except TypeError:
            raise unsupported_operand_error("+", type(self), type(other))

    @_auto_registration
    def __radd__(self, other) -> Composition:
        try:
            return pydec.add(other, self)
        except TypeError:
            raise unsupported_operand_error("+", type(other), type(self))

    @_auto_registration
    def __sub__(self, other) -> Composition:
        try:
            return pydec.sub(self, other)
        except TypeError:
            raise unsupported_operand_error("-", type(self), type(other))

    @_auto_registration
    def __rsub__(self, other) -> Composition:
        try:
            return pydec.sub(other, self)
        except TypeError:
            raise unsupported_operand_error("-", type(other), type(self))

    @_auto_registration
    def __isub__(self, other) -> Composition:
        if not isinstance(self, Composition):
            raise unsupported_operand_error("-=", type(self), type(other))
        if isinstance(other, Composition):
            return core.decBLAS.cc_sub_(self, other)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_sub_(self, other)
        else:
            raise unsupported_operand_error("-=", type(self), type(other))

    @_auto_registration
    def __matmul__(self, other) -> Composition:
        if isinstance(other, Tensor):
            return core.decBLAS.ct_matmul(self, other)
        else:
            raise unsupported_operand_error("@", type(self), type(other))

    @_auto_registration
    def __rmatmul__(self, other) -> Composition:
        if isinstance(other, Tensor):
            return core.decBLAS.tc_matmul(other, self)
        else:
            raise unsupported_operand_error("@=", type(other), type(other))

    @_auto_registration
    def __imul__(self, other) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("*=", type(self), type(other))
        if isinstance(other, Composition):
            return core.decMVF.cc_mul_(self, other)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_mul_(self, other)
        else:
            raise unsupported_operand_error("*=", type(self), type(other))

    @_auto_registration
    def __mul__(self, other) -> Composition:
        try:
            return pydec.mul(self, other)
        except TypeError:
            raise unsupported_operand_error("*", type(self), type(other))

    @_auto_registration
    def __rmul__(self, other) -> Composition:
        try:
            return pydec.mul(other, self)
        except TypeError:
            raise unsupported_operand_error("*", type(other), type(self))

    @_auto_registration
    def __itruediv__(self, other: Any) -> Composition:
        if not isinstance(self, Composition):
            raise unsupported_operand_error("/=", type(self), type(other))
        if isinstance(other, Composition):
            # TODO: not implement yet
            raise unsupported_operand_error("/=", type(self), type(other))
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_div_(self, other)
        else:
            raise unsupported_operand_error("/=", type(self), type(other))

    @_auto_registration
    def __truediv__(self, other: Any) -> Composition:
        try:
            return pydec.div(self, other)
        except TypeError:
            raise unsupported_operand_error("/", type(self), type(other))

    @_auto_registration
    def __rtruediv__(self, other: Any) -> Composition:
        try:
            return pydec.div(other, self)
        except TypeError:
            raise unsupported_operand_error("/", type(other), type(self))

    @_auto_registration
    def __eq__(self, other: Any) -> Tensor:  # type: ignore[override]
        if isinstance(other, Composition):
            return self.c_sum().__eq__(other.c_sum())
        return self.c_sum().__eq__(other)

    @_auto_registration
    def __ne__(self, other: Any) -> Tensor:  # type: ignore[override]
        if isinstance(other, Composition):
            return self.c_sum().__ne__(other.c_sum())
        return self.c_sum().__ne__(other)

    @_auto_registration
    def __gt__(self, other: Any) -> Tensor:
        if isinstance(other, Composition):
            return self.c_sum().__gt__(other.c_sum())
        return self.c_sum().__gt__(other)

    @_auto_registration
    def __lt__(self, other: Any) -> Tensor:
        if isinstance(other, Composition):
            return self.c_sum().__lt__(other.c_sum())
        return self.c_sum().__lt__(other)

    @_auto_registration
    def __ge__(self, other: Any) -> Tensor:
        if isinstance(other, Composition):
            return self.c_sum().__ge__(other.c_sum())
        return self.c_sum().__ge__(other)

    @_auto_registration
    def __le__(self, other: Any) -> Tensor:
        if isinstance(other, Composition):
            return self.c_sum().__le__(other.c_sum())
        return self.c_sum().__le__(other)

    @_auto_registration
    def add(
        self,
        other: Union[Composition, Tensor, Number],
        *,
        alpha: Optional[Number] = 1,
    ) -> Composition:
        """
        Note: although the function signature of `torch.Tensor.add` has parameter `out`, it is not working
        """
        return pydec.add(self, other, alpha=alpha)

    @_auto_registration
    def add_(
        self,
        other: Union[Composition, Tensor, Number],
        *,
        alpha: Optional[Number] = 1,
    ) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("add_", type(self), type(other))
        if isinstance(other, Composition):
            return core.decBLAS.cc_add_(self, other, alpha=alpha)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_add_(self, other, alpha=alpha)
        else:
            raise unsupported_operand_error("add_", type(self), type(other))

    @_auto_registration
    def sub(
        self,
        other: Union[Composition, Tensor, Number],
        *,
        alpha: Optional[Number] = 1,
    ) -> Composition:
        return pydec.sub(self, other, alpha=alpha)

    @_auto_registration
    def sub_(
        self, other: Union[Composition, Tensor, Number], *, alpha: Optional[Number] = 1
    ) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("sub_", type(self), type(other))
        if isinstance(other, Composition):
            return core.decBLAS.cc_sub_(self, other)
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_sub_(self, other)
        else:
            raise unsupported_operand_error("sub_", type(self), type(other))

    @_auto_registration
    def mul(self, other: Union[Tensor, Number]) -> Composition:
        return pydec.mul(self, other)

    @_auto_registration
    def mul_(self, other: Union[Tensor, Number]) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("mul_", type(self), type(other))
        if isinstance(other, Composition):
            # TODO: not implement yet
            raise unsupported_operand_error("mul_", type(self), type(other))
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_mul_(self, other)
        else:
            raise unsupported_operand_error("mul_", type(self), type(other))

    @_auto_registration
    def div(
        self, other: Union[Tensor, Number], *, rounding_mode: Optional[str] = None
    ) -> Composition:
        return pydec.div(self, other, rounding_mode=rounding_mode)

    @_auto_registration
    def div_(
        self, other: Union[Tensor, Number], *, rounding_mode: Optional[str] = None
    ) -> Composition:
        if not isinstance(self, Composition):
            # TODO: maybe return a composition is a good choice
            raise unsupported_operand_error("div_", type(self), type(other))
        if isinstance(other, Composition):
            # TODO: not implement yet
            raise unsupported_operand_error("div_", type(self), type(other))
        elif isinstance(other, (_int, _float, _bool, Tensor)):
            return core.decBLAS.ct_div_(self, other, rounding_mode=rounding_mode)
        else:
            raise unsupported_operand_error("div_", type(self), type(other))

    @_auto_registration
    def mv(self, vec: Tensor) -> Composition:
        return pydec.mv(self, vec)

    @_auto_registration
    def mm(self, mat2: Tensor) -> Composition:
        return pydec.mm(self, mat2)

    @overload
    def any(self) -> Tensor:
        ...

    @overload
    def any(self, dim: _int, keepdim: _bool = False) -> Tensor:
        ...

    @_auto_registration
    def any(self, *args: Any, **kwargs: Any) -> Tensor:
        return pydec.any(self, *args, **kwargs)

    @overload
    def all(self) -> Tensor:
        ...

    @overload
    def all(self, dim: _int, keepdim: _bool = False) -> Tensor:
        ...

    @_auto_registration
    def all(self, *args: Any, **kwargs: Any) -> Tensor:
        return pydec.all(self, *args, **kwargs)

    @overload
    def eq(self, other: Tensor) -> Composition:
        ...

    @overload
    def eq(self, other: Composition) -> Composition:
        ...

    @overload
    def eq(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def eq(self, other):
        return pydec.eq(self, other)

    @overload
    def ne(self, other: Tensor) -> Composition:
        ...

    @overload
    def ne(self, other: Composition) -> Composition:
        ...

    @overload
    def ne(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def ne(self, other):
        return pydec.ne(self, other)

    @overload
    def gt(self, other: Tensor) -> Composition:
        ...

    @overload
    def gt(self, other: Composition) -> Composition:
        ...

    @overload
    def gt(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def gt(self, other):
        return pydec.gt(self, other)

    @overload
    def lt(self, other: Tensor) -> Composition:
        ...

    @overload
    def lt(self, other: Composition) -> Composition:
        ...

    @overload
    def lt(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def lt(self, other):
        return pydec.lt(self, other)

    @overload
    def ge(self, other: Tensor) -> Composition:
        ...

    @overload
    def ge(self, other: Composition) -> Composition:
        ...

    @overload
    def ge(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def ge(self, other):
        return pydec.ge(self, other)

    @overload
    def le(self, other: Tensor) -> Composition:
        ...

    @overload
    def le(self, other: Composition) -> Composition:
        ...

    @overload
    def le(self, other: Number) -> Composition:
        ...

    @_auto_registration
    def le(self, other):
        return pydec.le(self, other)

    @_auto_registration
    def unsqueeze(self, dim: _int) -> Composition:
        return pydec.unsqueeze(self, dim)

    @overload
    def squeeze(self) -> Composition:
        ...

    @overload
    def squeeze(self, dim: _int) -> Composition:
        ...

    @_auto_registration
    def squeeze(self, dim=None) -> Composition:
        return pydec.squeeze(self, dim)

    @_auto_registration
    def unsqueeze_(self, dim: _int) -> Composition:
        self._residual_tensor.unsqueeze_(dim)
        self._component_tensor.unsqueeze_(_shift_dim(dim))
        return self

    @overload
    def squeeze_(self) -> Composition:
        ...

    @overload
    def squeeze_(self, dim: _int) -> Composition:
        ...

    @_auto_registration
    def squeeze_(self, dim=None) -> Composition:
        if dim is None:
            self._residual_tensor.squeeze_()
            if self.numc() == 1:
                self._component_tensor.squeeze_().unsqueeze_(0)
            else:
                self._component_tensor.squeeze_()
        else:
            self._residual_tensor.squeeze_(dim)
            self._component_tensor.squeeze_(_shift_dim(dim))
        return self

    @_auto_registration
    def transpose(self, dim0: _int, dim1: _int) -> Composition:
        return pydec.transpose(self, dim0, dim1)

    @_auto_registration
    def transpose_(self, dim0: _int, dim1: _int) -> Composition:
        self._residual_tensor.transpose_(dim0, dim1)
        self._component_tensor.transpose_(_shift_dim(dim0), _shift_dim(dim1))
        return self

    @overload
    def permute(self, dims: _size) -> Composition:
        ...

    @overload
    def permute(self, *dims: _int) -> Composition:
        ...

    @_auto_registration
    def permute(self, *args, **kwargs) -> Composition:
        if len(kwargs) == 1:
            dims = kwargs["dims"]
        elif isinstance(args[0], _int):
            dims = torch.Size(args)
        else:
            dims = args[0]
        return pydec.permute(self, dims=dims)

    @overload
    def sum(self, *, dtype: Optional[_dtype] = None) -> Composition:
        ...

    @overload
    def sum(
        self,
        dim: Union[_int, _size],
        keepdim: _bool = False,
        *,
        dtype: Optional[_dtype] = None,
    ) -> Composition:
        ...

    @_auto_registration
    def sum(
        self, dim=None, keepdim: _bool = False, *, dtype: Optional[_dtype] = None
    ) -> Composition:
        return pydec.sum(self, dim=dim, keepdim=keepdim, dtype=dtype)

    def c_sum(self, *, dtype: Optional[_dtype] = None) -> Tensor:
        return pydec.c_sum(self, dtype=dtype)

    @overload
    def mean(self, *, dtype: Optional[_dtype] = None) -> Composition:
        ...

    @overload
    def mean(
        self,
        dim: Union[_int, _size],
        keepdim: _bool = False,
        *,
        dtype: Optional[_dtype] = None,
    ) -> Composition:
        ...

    @_auto_registration
    def mean(
        self,
        dim=None,
        keepdim: _bool = False,
        *,
        dtype: Optional[_dtype] = None,
    ):
        return pydec.mean(self, dim, keepdim, dtype=dtype)

    @overload
    def view(self, dtype: _dtype) -> Composition:
        ...

    @overload
    def view(self, size: _size) -> Composition:
        ...

    @overload
    def view(self, *size: _int) -> Composition:
        ...

    @_auto_registration
    def view(self, *args, dtype=None, size=None) -> Composition:
        if dtype is None and size is None:
            if isinstance(args[0], _dtype):
                dtype = args[0]
            elif isinstance(args[0], _int):
                size = torch.Size(args)
            else:
                size = args[0]

        if dtype is not None:
            out_component_tensor = self._component_tensor.view(dtype)
            out_residual_tensor = self._residual_tensor.view(dtype)
        else:
            out_component_tensor = self._component_tensor.view((self.numc(),) + size)
            out_residual_tensor = self._residual_tensor.view(size)
        return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @overload
    def view_as(self, other: Tensor) -> Composition:
        ...

    @overload
    def view_as(self, other: Composition) -> Composition:
        ...

    @_auto_registration
    def view_as(self, other: Union[Tensor, Composition]) -> Composition:
        return self.view(other.size())

    @overload
    def reshape(self, shape: _size) -> Composition:
        ...

    @overload
    def reshape(self, *shape: _int) -> Composition:
        ...

    @_auto_registration
    def reshape(self, *args, shape=None) -> Composition:
        if shape is None:
            if isinstance(args[0], _int):
                shape = torch.Size(args)
            else:
                shape = args[0]
        return pydec.reshape(self, shape=shape)

    @_auto_registration
    def reshape_as(self, other: Tensor) -> Composition:
        return self.reshape(other.size())

    @_auto_registration
    def contiguous(self, memory_format=torch.contiguous_format) -> Composition:
        out_component_tensor = self._component_tensor.contiguous(
            memory_format=memory_format
        )
        out_residual_tensor = self._residual_tensor.contiguous(
            memory_format=memory_format
        )
        return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @_auto_registration
    def is_contiguous(self, memory_format=torch.contiguous_format) -> _bool:
        return self._component_tensor.is_contiguous(
            memory_format=memory_format
        ) and self._residual_tensor.is_contiguous(memory_format=memory_format)

    @overload
    def to(
        self, dtype: _dtype, non_blocking: _bool = False, copy: _bool = False
    ) -> Composition:
        ...

    @overload
    def to(
        self,
        device: Optional[Union[_device, str]] = None,
        dtype: Optional[_dtype] = None,
        non_blocking: _bool = False,
        copy: _bool = False,
    ) -> Composition:
        ...

    @overload
    def to(
        self,
        other: Tensor,
        non_blocking: _bool = False,
        copy: _bool = False,
    ) -> Composition:
        ...

    @overload
    def to(
        self,
        other: Composition,
        non_blocking: _bool = False,
        copy: _bool = False,
    ) -> Composition:
        ...

    @_auto_registration
    def to(self, *args, **kwargs) -> Composition:
        if isinstance(self, Tensor):
            assert isinstance(args[0], Composition)
            return self.to(args[0].residual, *args[1:], **kwargs)
        if isinstance(args[0], Composition):
            return self.to(args[0]._component_tensor, *args[1:], **kwargs)
        else:
            out_component_tensor = self._component_tensor.to(*args, **kwargs)
            out_residual_tensor = self._residual_tensor.to(*args, **kwargs)
            return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @overload
    def masked_fill(self, mask: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def masked_fill(self, mask: Tensor, value: Number) -> Composition:
        ...

    @_auto_registration
    def masked_fill(self, mask: Tensor, value: Any) -> Composition:
        return pydec.masked_fill(self, mask, value)

    @overload
    def masked_fill_(self, mask: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def masked_fill_(self, mask: Tensor, value: Number) -> Composition:
        ...

    @_auto_registration
    def masked_fill_(self, mask: Tensor, value: Any) -> Composition:
        r"""
        Unsafe.
        """
        self._component_tensor.masked_fill_(mask[None], value)
        self._residual_tensor.masked_fill_(mask, value)
        return self

    @overload
    def c_masked_fill(self, mask: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def c_masked_fill(self, mask: Tensor, value: Number) -> Composition:
        ...

    def c_masked_fill(self, mask: Tensor, value: Any) -> Composition:
        return pydec.c_masked_fill(self, mask, value)

    @overload
    def c_masked_fill_(self, mask: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def c_masked_fill_(self, mask: Tensor, value: Number) -> Composition:
        ...

    def c_masked_fill_(self, mask: Tensor, value: Any) -> Composition:
        if mask.dim() == 1:
            if len(mask) != self.numc():
                raise arg_value_error(
                    f"the length of mask ({len(mask)}) should match component number ({self.numc()})"
                )
            mask_size = (self.numc(),) + (1,) * self.dim()
            mask = mask.view(mask_size)
        self._component_tensor.masked_fill_(mask, value)
        return self

    @_auto_registration
    def masked_select(self, mask: Tensor) -> Composition:
        return pydec.masked_select(self, mask)

    @_auto_registration
    def masked_scatter(self, mask: Tensor, source: Tensor) -> Composition:
        return pydec.masked_scatter(self, mask, source)

    @_auto_registration
    def masked_scatter_(self, mask: Tensor, source: Tensor) -> Composition:
        # TODO: seems not make sense, see examples in docs
        self._component_tensor.masked_scatter_(mask[None], source)
        self._residual_tensor.masked_scatter_(mask, source)
        return self

    @_auto_registration
    def gather(
        self, dim: _int, index: Tensor, *, sparse_grad: _bool = False
    ) -> Composition:
        return pydec.gather(self, dim, index, sparse_grad=sparse_grad)

    @overload
    def scatter(self, dim: _int, index: Tensor, src: Tensor) -> Composition:
        ...

    @overload
    def scatter(
        self, dim: _int, index: Tensor, src: Tensor, *, reduce: str
    ) -> Composition:
        ...

    @overload
    def scatter(
        self, dim: _int, index: Tensor, value: Number, *, reduce: str
    ) -> Composition:
        ...

    @overload
    def scatter(self, dim: _int, index: Tensor, value: Number) -> Composition:
        ...

    @_auto_registration
    def scatter(
        self,
        dim: _int,
        index: Tensor,
        src_value: Optional[Union[Tensor, Number]] = None,
        *,
        reduce: Optional[str] = None,
        src: Optional[Tensor] = None,
        value: Optional[Number] = None,
    ) -> Composition:
        if src_value is None:
            src_value = src if src is not None else value
        return pydec.scatter(self, dim, index, src_value, reduce=reduce)  # type: ignore[arg-type]

    @overload
    def scatter_(self, dim: _int, index: Tensor, src: Tensor) -> Tensor:
        ...

    @overload
    def scatter_(self, dim: _int, index: Tensor, src: Tensor, *, reduce: str) -> Tensor:
        ...

    @overload
    def scatter_(
        self, dim: _int, index: Tensor, value: Number, *, reduce: str
    ) -> Tensor:
        ...

    @overload
    def scatter_(self, dim: _int, index: Tensor, value: Number) -> Tensor:
        ...

    @_auto_registration
    def scatter_(
        self, dim: Any, index: Tensor, src: Any, *, reduce: str = None
    ) -> Composition:
        if reduce == "add":
            holder = torch.zeros_like(self._residual_tensor).to(self._residual_tensor)
            holder = holder.scatter(dim, index, src, reduce=reduce)
            self += holder
            return self
        else:
            c_index = index[None].expand((self.numc(),) + (-1,) * index.dim())
            if isinstance(src, Tensor):
                c_src = src[None].expand((self.numc(),) + (-1,) * src.dim())
            else:
                c_src = src
            if reduce is None:
                self._component_tensor.scatter_(_shift_dim(dim), c_index, c_src)
                self._residual_tensor.scatter_(dim, index, src)
            else:
                self._component_tensor.scatter_(
                    _shift_dim(dim), c_index, c_src, reduce=reduce
                )
                self._residual_tensor.scatter_(dim, index, src, reduce=reduce)
            return self

    @_auto_registration
    def diagonal_scatter(
        self, src: Tensor, offset: _int = 0, dim1: _int = 0, dim2: _int = 1
    ) -> Composition:
        return pydec.diagonal_scatter(self, src, offset, dim1, dim2)

    @_auto_registration
    def cuda(
        self,
        device: Optional[Union[_device, _int, str]] = None,
        non_blocking: _bool = False,
    ) -> Composition:
        out_component_tensor = self._component_tensor.cuda(
            device=device, non_blocking=non_blocking
        )
        out_residual_tensor = self._residual_tensor.cuda(
            device=device, non_blocking=non_blocking
        )
        return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @_auto_registration
    def cpu(self) -> Composition:
        out_component_tensor = self._component_tensor.cpu()
        out_residual_tensor = self._residual_tensor.cpu()
        return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @overload
    def index_select(self, dim: _int, index: Tensor) -> Composition:
        ...

    @_auto_registration
    def index_select(self, dim: _int, index: Tensor) -> Composition:
        return pydec.index_select(self, dim=dim, index=index)

    @overload
    def c_index_select(self, index: Tensor, with_residual: _bool = True) -> Composition:
        ...

    def c_index_select(self, index: Tensor, with_residual: _bool = True) -> Composition:
        return pydec.c_index_select(self, index=index, with_residual=with_residual)

    @_auto_registration
    def masked_select(self, mask: Tensor) -> Composition:
        return pydec.masked_select(self, mask)

    @overload
    def index_fill(self, dim: _int, index: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def index_fill(self, dim: _int, index: Tensor, value: Number) -> Composition:
        ...

    @_auto_registration
    def index_fill(self, dim: _int, index: Tensor, value: Any) -> Composition:
        return pydec.index_fill(self, dim=dim, index=index, value=value)

    @overload
    def index_fill_(self, dim: _int, index: Tensor, value: Tensor) -> Composition:
        ...

    @overload
    def index_fill_(self, dim: _int, index: Tensor, value: Number) -> Composition:
        ...

    @_auto_registration
    def index_fill_(self, dim: _int, index: Tensor, value: Number) -> Composition:
        self._component_tensor.index_fill_(
            dim=_shift_dim(dim), index=index, value=value
        )
        self._residual_tensor.index_fill_(dim=dim, index=index, value=value)
        return self

    @overload
    def select(self, dim: _int, index: _int) -> Composition:
        ...

    @_auto_registration
    def select(self, dim: _int, index: _int) -> Composition:
        return pydec.select(self, dim=dim, index=index)

    @overload
    def type(self, dtype: None = None, non_blocking: _bool = False) -> str:
        ...

    @overload
    def type(self, dtype: Union[str, _dtype], non_blocking: _bool = False) -> Tensor:
        ...

    @_auto_registration
    def type(self, dtype=None, non_blocking: _bool = False):
        if dtype is None:
            return self._residual_tensor.type()
        else:
            out_component_tensor = self._component_tensor.type(
                dtype=dtype, non_blocking=non_blocking
            )
            out_residual_tensor = self._residual_tensor.type(
                dtype=dtype, non_blocking=non_blocking
            )
            return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @overload
    def type_as(self, other: Tensor) -> Composition:
        ...

    @overload
    def type_as(self, other: Composition) -> Composition:
        ...

    @_auto_registration
    def type_as(self, other: Union[Tensor, Composition]) -> Composition:
        if isinstance(other, Composition):
            out_component_tensor = self._component_tensor.type_as(
                other._component_tensor
            )
            out_residual_tensor = self._residual_tensor.type_as(other._component_tensor)
        else:
            out_component_tensor = self._component_tensor.type_as(other)
            out_residual_tensor = self._residual_tensor.type_as(other)
        return pydec.as_composition(out_component_tensor, out_residual_tensor)

    @overload
    def round(self) -> Composition:
        ...

    @overload
    def round(self, *, decimals: _int) -> Composition:
        ...

    @_auto_registration
    def round(self, *, decimals: _int = None):
        return pydec.round(self, decimals=decimals)

    @overload
    def round_(self) -> Composition:
        ...

    @overload
    def round_(self, *, decimals: _int) -> Composition:
        ...

    @_auto_registration
    def round_(self, *, decimals: _int = None) -> Composition:
        return pydec.round_(self, decimals=decimals)

    @_auto_registration
    def abs(self) -> Composition:
        return pydec.abs(self)

    @_auto_registration
    def abs_(self) -> Composition:
        return pydec.abs_(self)

    @_auto_registration
    def requires_grad_(self, mode: _bool = True) -> Composition:
        self._component_tensor.requires_grad_(mode)
        self._residual_tensor.requires_grad_(mode)
        return self

    @_auto_registration
    def apply_(self, callable: Callable) -> Composition:
        self._component_tensor.apply_(callable)
        self._residual_tensor.apply_(callable)
        return self

    @_auto_registration
    def map_(self, composition: Composition, callable: Callable) -> Composition:
        self._residual_tensor.map_(composition._residual_tensor, callable)
        # permute to be broadcastable
        p_dims = [i for i in range(1, self._component_tensor.dim())] + [0]
        p_component_tensor = self._component_tensor.permute(*p_dims)
        p_dims = [i for i in range(1, composition._component_tensor.dim())] + [0]
        p_other_component_tensor = composition.permute(*p_dims)
        p_component_tensor.map_(p_other_component_tensor, callable)

        # recover
        p_dims = [-1] + [i for i in range(0, p_component_tensor.dim() - 1)]
        self._component_tensor = p_component_tensor.permute(p_dims)
        return self

    @_auto_registration
    def relu(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.relu(self, ref=ref)

    @_auto_registration
    def relu_(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.relu_(self, ref=ref)

    @_auto_registration
    def tanh(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.tanh(self, ref=ref)

    @_auto_registration
    def tanh_(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.tanh_(self, ref=ref)

    @_auto_registration
    def sigmoid(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.sigmoid(self, ref=ref)

    @_auto_registration
    def sigmoid_(self, *, ref: Optional[Tensor] = None) -> Composition:
        return pydec.sigmoid_(self, ref=ref)

    @_auto_registration
    def reciprocal(self) -> Composition:
        return pydec.reciprocal(self)

    @_auto_registration
    def reciprocal_(self) -> Composition:
        return pydec.reciprocal_(self)

    @_auto_registration
    def exp(self) -> Composition:
        return pydec.exp(self)

    @_auto_registration
    def exp(self) -> Composition:
        return pydec.exp_(self)

    @overload
    def softmax(self, dim: _int, dtype: Optional[_dtype] = None) -> Composition:
        ...

    @overload
    def softmax(
        self, dim: Union[str, ellipsis, None], *, dtype: Optional[_dtype] = None
    ) -> Composition:
        ...

    @_auto_registration
    def softmax(self, dim: Any, dtype: Optional[_dtype] = None) -> Composition:
        return pydec.softmax(self, dim, dtype)

    def sqrt(self) -> Composition:
        return pydec.sqrt(self)

    def sqrt_(self) -> Composition:
        return pydec.sqrt_(self)


class IndexComposition(Composition):
    """
    TODO
    """

    # we use this number to represent empty value
    # all indices should avoid this value
    MASK_NUM = -2147483648

    @property
    def empty_mask(self) -> Tensor:
        return (
            self.components == IndexComposition.MASK_NUM,
            self.residual == IndexComposition.MASK_NUM,
        )

    @overload
    def __init__(
        self,
        component_tensor: Tensor,
        residual_tensor: Tensor = None,
    ) -> None:
        ...

    @overload
    def __init__(self, composition: IndexComposition) -> None:
        ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if len(args) + len(kwargs) == 0:
            # a private constructor to create a void composition with no data,
            # which is usually assigned in `_from_replace`.
            return

        # Type checking
        accept_dtypes = [torch.int, torch.long]
        msg = "Expected arguments to have one of the following scalar types: {}; but got {} instead (while checking arguments for IndexComposition)"

        if self.dtype not in accept_dtypes:
            raise RuntimeError(msg.format(accept_dtypes, self.dtype))

    def __repr__(self, *, composition_contents: List[str] = None) -> str:
        # TODO
        # TODO: add example for docs
        return super().__repr__(composition_contents=composition_contents)


class _C_AccessingMode:
    is_enabled = False

    @classmethod
    def set_enabled(cls, enabled: _bool):
        cls.is_enabled = enabled


def is_c_accessing_enabled() -> _bool:
    return _C_AccessingMode.is_enabled


class enable_c_accessing(_DecoratorContextManager):
    r"""Context-manager that enables component indexing and slicing."""

    def __enter__(self) -> None:
        self.prev = is_c_accessing_enabled()
        _C_AccessingMode.set_enabled(True)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _C_AccessingMode.set_enabled(self.prev)


class set_c_accessing_enabled(_DecoratorContextManager):
    r"""Context-manager that sets c_accessing to on or off."""

    def __init__(self, mode: bool) -> None:
        self.prev = is_c_accessing_enabled()
        _C_AccessingMode.set_enabled(mode)
        self.mode = mode

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        _C_AccessingMode.set_enabled(self.prev)

    def clone(self):
        return self.__class__(self.mode)


class _C_AccessingComposition(Composition):
    r"""
    TODO: A temporary type for implementation of component subscript accessing for Composition.
    """

    def __init__(self, composition: Composition):
        super().__init__()  # void initialization
        self._component_tensor = composition._component_tensor
        self._residual_tensor = composition._residual_tensor

    def __getitem__(
        self, indices: Union[None, _int, slice, Tensor, List, Tuple]
    ) -> Union[Composition, Tensor]:
        return super().__c_getitem__(indices)

    def __setitem__(
        self,
        indices: Union[None, _int, slice, Tensor, List, Tuple],
        val: Union[Composition, Tensor, Number],
    ) -> None:
        return super().__c_setitem__(indices, val)

    def __len__(self):
        return self.components.__len__()

    def __iter__(self):
        return self._component_tensor.__iter__()



import torch
from typing import (
    Any,
    Union,
    List,
    Dict,
    Tuple,
    Optional,
    Callable,
    overload,
    Sequence,
    Iterable,
)

# In some cases, these basic types are shadowed by corresponding
# top-level values.  The underscore variants let us refer to these
# types.  See https://github.com/python/mypy/issues/4146 for why these
# workarounds is necessary
# In the future, this will become useful if mypy is introduced into pydec
from torch.types import (
    _int,
    _float,
    _bool,
    Number,
    _dtype,
    _device,
    _qscheme,
    _size,
    _layout,
)


def _shift_dim(dim: _int) -> _int:
    if dim >= 0:
        dim += 1
    return dim


@overload
def _shift_dims(dims: _size, /) -> torch.Size:
    ...


@overload
def _shift_dims(*dims: _int) -> torch.Size:
    ...


def _shift_dims(*dims: Any) -> torch.Size:
    if len(dims) == 1:
        dims = dims[0]
    out = list(dims)
    for i in range(len(dims)):
        out[i] = _shift_dim(dims[i])
    return torch.Size(out)


def parse_args(args: Iterable[Any], key_list: Iterable[str], kwargs: Dict[str, Any]):
    for key, arg in zip(key_list, args):
        kwargs[key] = arg
